/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.lib.misc;

import javax.annotation.Nonnull;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumFacing.Axis;
import net.minecraft.util.EnumFacing.AxisDirection;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3d;
import net.minecraft.util.math.Vec3i;

/** Class for dealing with {@link Vec3d}, {@link Vec3i}, {@link EnumFacing}, {@link Axis} conversions and additions.
 * This is for simple functions ONLY, {@link PositionUtil} is for complex interactions */
public class VecUtil {
    public static final BlockPos POS_ONE = new BlockPos(1, 1, 1);
    public static final Vec3d VEC_HALF = new Vec3d(0.5, 0.5, 0.5);
    public static final Vec3d VEC_ONE = new Vec3d(1, 1, 1);

    public static Vec3d add(Vec3d a, Vec3i b) {
        return a.func_72441_c(b.func_177958_n(), b.func_177956_o(), b.func_177952_p());
    }

    public static Vec3d offset(Vec3d from, EnumFacing direction, double by) {
        return from.func_72441_c(direction.func_82601_c() * by, direction.func_96559_d() * by, direction.func_82599_e() * by);
    }

    public static double dot(Vec3d a, Vec3d b) {
        return a.field_72450_a * b.field_72450_a + a.field_72448_b * b.field_72448_b + a.field_72449_c * b.field_72449_c;
    }

    public static Vec3d scale(Vec3d vec, double scale) {
        return vec.func_186678_a(scale);
    }

    public static EnumFacing getFacing(Axis axis, boolean positive) {
        AxisDirection dir = positive ? AxisDirection.POSITIVE : AxisDirection.NEGATIVE;
        return EnumFacing.func_181076_a(dir, axis);
    }

    public static BlockPos absolute(BlockPos val) {
        return new BlockPos(Math.abs(val.func_177958_n()), Math.abs(val.func_177956_o()), Math.abs(val.func_177952_p()));
    }

    public static Vec3d replaceValue(Vec3d old, Axis axis, double with) {
        return new Vec3d(//
            axis == Axis.X ? with : old.field_72450_a,//
            axis == Axis.Y ? with : old.field_72448_b,//
            axis == Axis.Z ? with : old.field_72449_c//
        );
    }

    @Nonnull
    public static BlockPos replaceValue(Vec3i old, Axis axis, int with) {
        return new BlockPos(//
            axis == Axis.X ? with : old.func_177958_n(),//
            axis == Axis.Y ? with : old.func_177956_o(),//
            axis == Axis.Z ? with : old.func_177952_p()//
        );
    }

    public static double getValue(Vec3d from, Axis axis) {
        return axis == Axis.X ? from.field_72450_a : axis == Axis.Y ? from.field_72448_b : from.field_72449_c;
    }

    public static int getValue(Vec3i from, Axis axis) {
        return axis == Axis.X ? from.func_177958_n() : axis == Axis.Y ? from.func_177956_o() : from.func_177952_p();
    }

    public static double getValue(Vec3d negative, Vec3d positive, EnumFacing face) {
        switch (face) {
            case DOWN:
                return negative.field_72448_b;
            case UP:
                return positive.field_72448_b;
            case NORTH:
                return negative.field_72449_c;
            case SOUTH:
                return positive.field_72449_c;
            case WEST:
                return negative.field_72450_a;
            case EAST:
                return positive.field_72450_a;
            default:
                throw new IllegalArgumentException("Unknwon EnumFacing " + face);
        }
    }

    public static int getValue(Vec3i negative, Vec3i positive, EnumFacing face) {
        switch (face) {
            case DOWN:
                return negative.func_177956_o();
            case UP:
                return positive.func_177956_o();
            case NORTH:
                return negative.func_177952_p();
            case SOUTH:
                return positive.func_177952_p();
            case WEST:
                return negative.func_177958_n();
            case EAST:
                return positive.func_177958_n();
            default:
                throw new IllegalArgumentException("Unknwon EnumFacing " + face);
        }
    }

    public static Vec3d convertCenter(Vec3i pos) {
        return new Vec3d(pos.func_177958_n() + 0.5, pos.func_177956_o() + 0.5, pos.func_177952_p() + 0.5);
    }

    public static BlockPos convertFloor(Vec3d vec) {
        return new BlockPos(Math.floor(vec.field_72450_a), Math.floor(vec.field_72448_b), Math.floor(vec.field_72449_c));
    }

    public static BlockPos convertCeiling(Vec3d vec) {
        return new BlockPos(Math.ceil(vec.field_72450_a), Math.ceil(vec.field_72448_b), Math.ceil(vec.field_72449_c));
    }

    public static Tuple3f convertFloat(Vec3d vec) {
        return new Vector3f((float) vec.field_72450_a, (float) vec.field_72448_b, (float) vec.field_72449_c);
    }

    // Min/Max

    public static BlockPos min(BlockPos a, BlockPos b) {
        if (a == null) return b;
        if (b == null) return a;
        return new BlockPos(//
            Math.min(a.func_177958_n(), b.func_177958_n()),//
            Math.min(a.func_177956_o(), b.func_177956_o()),//
            Math.min(a.func_177952_p(), b.func_177952_p())//
        );
    }

    public static BlockPos min(BlockPos a, BlockPos b, BlockPos c) {
        return min(min(a, b), c);
    }

    public static BlockPos min(BlockPos a, BlockPos b, BlockPos c, BlockPos d) {
        return min(min(a, b), min(c, d));
    }

    public static BlockPos max(BlockPos a, BlockPos b) {
        if (a == null) return b;
        if (b == null) return a;
        return new BlockPos(//
            Math.max(a.func_177958_n(), b.func_177958_n()),//
            Math.max(a.func_177956_o(), b.func_177956_o()),//
            Math.max(a.func_177952_p(), b.func_177952_p())//
        );
    }

    public static BlockPos max(BlockPos a, BlockPos b, BlockPos c) {
        return max(max(a, b), c);
    }

    public static BlockPos max(BlockPos a, BlockPos b, BlockPos c, BlockPos d) {
        return max(max(a, b), max(c, d));
    }

    public static Vec3d min(Vec3d a, Vec3d b) {
        if (a == null) return b;
        if (b == null) return a;
        return new Vec3d(//
            Math.min(a.field_72450_a, b.field_72450_a),//
            Math.min(a.field_72448_b, b.field_72448_b),//
            Math.min(a.field_72449_c, b.field_72449_c)//
        );
    }

    public static Vec3d min(Vec3d a, Vec3d b, Vec3d c) {
        return min(min(a, b), c);
    }

    public static Vec3d min(Vec3d a, Vec3d b, Vec3d c, Vec3d d) {
        return min(min(a, b), min(c, d));
    }

    public static Vec3d max(Vec3d a, Vec3d b) {
        if (a == null) return b;
        if (b == null) return a;
        return new Vec3d(//
            Math.max(a.field_72450_a, b.field_72450_a),//
            Math.max(a.field_72448_b, b.field_72448_b),//
            Math.max(a.field_72449_c, b.field_72449_c)//
        );
    }

    public static Vec3d max(Vec3d a, Vec3d b, Vec3d c) {
        return max(max(a, b), c);
    }

    public static Vec3d max(Vec3d a, Vec3d b, Vec3d c, Vec3d d) {
        return max(max(a, b), max(c, d));
    }
}
